;;;
;;; circuit-bus.lisp
;;;
;;; support multi-wire busses in circuit-sim.lisp
;;; 
;;;
;;; $Revision: 1.3 $
;;;
;;; $Log: circuit-bus.lisp,v $
;;; Revision 1.3  2007/02/11 18:45:13  jao
;;; 1. In package "CIRCUIT"
;;; 2. Move gated-edge probe to circuit-sim.lisp
;;;
;;; Revision 1.2  2007/02/10 23:48:31  jao
;;; 1. get-integer-signal
;;; 2. Optional integer format argument for probe-integer
;;; 3. gated-edge-integer-probe, gated-edge-probe
;;;
;;; Revision 1.1  2003/10/11 03:10:38  jao
;;; Initial revision
;;;
;;;

;;; Busses are meant to represent multiple wires as a single "signal"
;;; in the circuit.
;;;
;;; For now, busses are represented as lists of the component wires.
;;;
;;; I have a desire to represent busses as integers, but what happens
;;; when one wire in a bus has a bogus value? I need some fallback
;;; mechanism for non-binary values in a bus.
;;;

(in-package "CIRCUIT")

(defclass bus (wire) ())

(defun make-bus (&rest wires)
  "Creates a bus which has WIRES as its component signals. 
When treated numerically, the first WIRE is the least-significant bit."
  (let ((b (make-instance 'bus :signal-value
			  wires)))
    (mapcar (lambda (subwire) 
	      (push #'(lambda () (mapcar #'funcall (change-actions b)))
		    (change-actions subwire)))
	    wires)
    b))

(defmethod bus-width ((w wire))
  1)

(defmethod bus-width ((b bus))
  "Returns the number of components making up the bus. This of course 
assumes all of the bus components are individual bits, and not busses 
themselves."
  (length (signal-value b)))


(defmethod iterate-bus (mapping-function (b bus))
  "Iterates over the bus, providing mapping function with each wire
in turn. I.e. calls (mapping-function wire) for each wire, starting
with LSB, progressing to MSB."
  (mapcar mapping-function (signal-value b)))

(defmethod iterate-bus-counting (mapping-function (b bus))
  "Iterates over the bus, providing mapping function with each wire
and the bit number in turn. I.e. calls (mapping-function wire bit-num)
for each wire, starting with LSB, 0 and progressing to MSB, N-1"
  (let ((i 0))
    (mapcar #'(lambda (w)
		(prog1 
		    (funcall mapping-function w i)
		  (incf i)))
	    (signal-value b))))
			     
(defmethod set-signal ((b bus) new-values)
  (mapcar #'set-signal (signal-value b) 
	  (if (integerp new-values)
	      (let ((bitlist nil))
		(dotimes (bit (length (signal-value b)) (nreverse bitlist))
		  (push (ldb (byte 1 bit) new-values) bitlist)))
	      new-values)))

(defmethod get-signal ((b bus))
  (mapcar #'get-signal (signal-value b)))

(defmethod get-integer-signal ((w wire))
  (let ((s (get-signal w)))
    (if (logical-valid s)
	(if (logical-false s)
	    0
	    1)
	s)))

(defmethod get-integer-signal ((b bus))
  "Encode B's value into an integer. If any signal is not a logical value, 
returns that value."
  (let ((num 0) (bit 0) (siglist (get-signal b)))
    (dolist (s siglist num)
      (setf (ldb (byte 1 bit) num)
	    (if (logical-true s)
		1
		(if (logical-false s)
		    0
		    (return-from get-integer-signal s))))
      (incf bit))))

(defmethod nth-wire (n (b bus))		  
  "Returns the Nth element of the bus. N=0 is the LSB when bus values
are considered as an integer."
  (nth n (signal-value b)))

(defmethod probe-integer ((b bus) name &optional (integer-format "~D"))
  (add-action b #'(lambda () (format t 
				     (format nil 
					     "~~A Time: ~~D New Value: ~A~~%"
					     integer-format)
				     name (agenda-time *agenda*)
                                     (get-integer-signal b)))))

;; to add action to a bus, add an action to all component wires to 
;; trigger an action on the bus.

(defmethod add-action ((b bus) func &optional (output-initializations nil))
    "OUTPUT-INITIALIZATIONS should be either NIL or a list of functions without arguments. 
If NIL, the action FUNC is called once immediately, to initialize outputs. For
sequential circuits, that may not make sense. If non-NIL, each function in the
OUTPUT-INITIALIZATIONS list is called in lieu of calling FUNC."

  (push func (change-actions b))
  (if (null output-initializations)
    (funcall func) ;; must call to ensure outputs assume correct state for initial inputs
    (mapcar #'funcall output-initializations)))

(defmethod make-gated-edge-integer-probe ((b bus) (edge-wire wire) name
					  &optional (edge-type :rising)
					  (integer-format "~D"))
				  
 "Makes a gated probe which samples B when EDGE-WIRE has an edge transition
that matches EDGE-TYPE (:RISING or :FALLING). Should add setup & hold time
requirements on this probe, but for now, close enough."

  (let ((probe (make-instance 'gated-probe :enabled nil)))
    (add-action edge-wire
		(ecase edge-type
		  (:rising
		   #'(lambda ()
		       (when (logical-true (get-signal edge-wire))
			 (when (enabled probe)
			   (format t 
				   (format nil "~~A Time: ~~D New Value: ~A~~%"
					   integer-format)
				   name (agenda-time *agenda*)
				   (get-integer-signal b))))))
		   
		  (:falling
		   #'(lambda ()
		       (when (logical-false (get-signal edge-wire))
			 (when (enabled probe)
			   (format t 
				   (format nil "~~A Time: ~~D New Value: ~A~~%"
					   integer-format)
				   name (agenda-time *agenda*)
				   (get-integer-signal b))))))))
    probe))

