;;;
;;; circuit-clock.lisp
;;;
;;; An attempt to model clock signals and support limited-time
;;; simulation for circuit-sim.lisp
;;;
;;; $Revision: 1.2 $
;;;
;;; $Log: circuit-clock.lisp,v $
;;; Revision 1.2  2007/02/11 18:45:40  jao
;;; In package "CIRCUIT"
;;;
;;; Revision 1.1  2007/02/11 00:38:47  jao
;;; Initial revision
;;;
;;;

(in-package "CIRCUIT")

;;; The basic clock is a square wave with a certain initial rising (falling)
;;; edge, a time spent high (low), and a time spent low (high), ad
;;; infinitum. We can model this quite simply by pushing the next clock edge 
;;; onto the *agenda* at each edge. However, the naive version of "propagate"
;;; will not work properly in the presence of this because it will continue
;;; working until the agenda is empty. 

(defun make-clock (wire initial-edge-time initial-edge-type t0 t1)
  "Causes a 'spontaneous' clock waveform to appear on WIRE, beginning
at INITIAL-EDGE-TIME, which should be equal to or later than the 
current simulation time. The INITIAL-EDGE-TYPE should be :rising or 
:falling. t0 is the duration of the first interval (high for :rising, 
low for :falling), and t1 the duration of the second interval (low for
:rising, high for :falling.

  Do not call 'propagate' in the presence of a clock unless you have
some way to stop the infinite loop..."
  (let ((now (agenda-time *agenda*)))
    (when (< initial-edge-time now)
      (error "Clock cannot begin in the past."))
    (let ((initial-delay (- initial-edge-time now))
	  first-level
	  second-level)
      (ecase initial-edge-type
	(:rising (setf first-level 1 second-level 0))
	(:falling (setf first-level 0 second-level 1)))
      (labels ((first-event ()
		 "implements initial edge"
		 (after-delay t1 (set-signal wire first-level)))
	       (second-event ()
		 (after-delay t0 (set-signal wire second-level)))
	       
	       (process-clock-event ()
		 (let ((current-level (get-signal wire)))
		   (cond
		     ((logical-eqv current-level first-level)
		      (second-event))
		     ((logical-eqv current-level second-level)
		      (first-event))
		     (t (error "Clock signal disrupted."))))))

	(add-action wire #'process-clock-event 
		    (list #'(lambda ()
			      (after-delay initial-delay 
					   (set-signal 
					    wire 
					    first-level)))))))))

(defun propagate-until (end-time)
  "Runs events on the agenda until end-time is reached, or the agenda is 
empty. Events at end-time are executed."
  (do ((item))
      ((or (empty *agenda*) (> (agenda-time *agenda*)
			       end-time)) (agenda-time *agenda*))
    (setf item (first-item *agenda*))
    (funcall item)
    (remove-first-item *agenda*)))