;;;
;;; circuit-logic.lisp
;;;
;;; Helper routines for circuit-sim.lisp which represent 
;;; logical states.
;;;
;;; $Revision: 1.2 $
;;;
;;; $Log: circuit-logic.lisp,v $
;;; Revision 1.2  2007/02/11 18:49:20  jao
;;; 1. in package "CIRCUIT"
;;; 2. logical-nor
;;;
;;; Revision 1.1  2007/02/11 00:39:34  jao
;;; Initial revision
;;;
;;;

(in-package "CIRCUIT")

(defun logical-true (value)
  "Attempts to provide a notion of 'truth' which is useful for
logical circuits. True values are 1, t. Other values are not true.
They might not be false, because we wish values such as :invalid
not to be definitely true or definitely false."
  (or (and (numberp value) (= value 1))
      (eq value t)))

(defun logical-false (value)
  "Attempts to provide a notion of 'untruth' which is useful for
logical circuits. False values are 0, nil. Other values are not false.
They might not be true, because we wish values such as :invalid
not to be definitely true or definitely false."
  (or (and (numberp value) (zerop value))
      (not value)))

(defun logical-valid (value)
  "T if value is a definite logical value. Nil otherwise."
  (or (logical-true value)
      (logical-false value)))

(defun logical-eqv (v1 v2)
  "T if v1, v2 are logically equivalent (both true or both false), 
Nil otherwise."
  (or (and (logical-true v1) (logical-true v2))
      (and (logical-false v1) (logical-false v2))))

(defun logical-not (value)
  (cond 
    ((logical-true value) 0)
    ((logical-false value) 1)
    (t value)))

(defun logical-and (&rest values)
  "True if all values are logically true. Otherwise, nil, 
unless all logically valid values are true, which gives 
an invalid value the chance to influence, hence, invalidate
the result."
  ;;; somehow, should be able to search only once through the list...
  (cond
    ((every #'logical-true values) t)
    ((notany #'logical-false values) (find-if-not #'logical-valid values))))

(defun logical-nand (&rest values)
  (logical-not (apply #'logical-and values)))

(defun logical-or (&rest values)
  "True if any values are logically true. Otherwise,
logically false, unless an invalid input invalidates the output."
  ;;; somehow, should be able to search only once through the list
  (let ((decisive (find-if #'logical-true values)))
    (if decisive
	decisive
	(find-if-not #'logical-valid values))))

(defun logical-nor (&rest values)
  (logical-not (apply #'logical-or values)))

;;;
;;; ideally, the logical-or and logical-and operators should
;;; preserve the "flavor" of the inputs. I.e. 0 0 0 should produce 0, 
;;; nil nil nil should produce nil. 
;;;

;;; perhaps should come up with some way to test against "truth tables."
;;;
;;; i.e. specify logic functions as
;;;
;;; e.g.: binary NAND  (1 1) -> 0
;;;                    (0 1) -> 1
;;;                    (1 0) -> 1
;;;                    (0 0) -> 1
;;;
;;; where any "logical true" value matches a logical true input,
;;; logical false matches a logical false input. 
;;;
;;; also, should allow "don't cares"
;;;
;;;  binary NAND (0 :X) -> 1
;;;              (:X 0) -> 1
;;;              (1  1) -> 0
;;;
;;; or "valid" :V? 
;;; 
      