;;;
;;; circuit-sim.lisp
;;;
;;; An attempt to implement the digital circuit simulation 
;;; model from SICP Section 3.3.4
;;;
;;; Using Common Lisp objects to replace the Scheme'ish 
;;; dispatch mechanism.
;;;
;;;
;;; $Revision: 1.4 $
;;;
;;; $Log: circuit-sim.lisp,v $
;;; Revision 1.4  2007/02/11 19:27:05  jao
;;; Export accumulate-probe
;;;
;;; Revision 1.3  2007/02/11 18:48:27  jao
;;; 1. Add package "CIRCUIT" definition.
;;; 2. at-time-fun, accumulate probe
;;; 3. gated-probe, gated-edge-probe moved here
;;;
;;; Revision 1.2  2007/02/10 23:50:52  jao
;;; 1. Fix to after-delay
;;; 2. Fix probe to use get-signal instead of signal-value.
;;;
;;; Revision 1.1  2003/10/11 03:10:18  jao
;;; Initial revision
;;;
;;;

;;; depends on: package.lisp, queue.lisp

(in-package "CIRCUIT")

;;; WIRE
;;;
;;; The basic signal propagation mechanism. 
;;; Wires have a value (could be an integer (multi-wire bus) or
;;; :invalid or some such)
;;;
;;; When the value changes, it fires a series of change actions.
;;;
;;; In future revision, this should contain some memory of how long
;;; the wire has been valid, in order to verify that setup-and-hold
;;; requirements are met on latches, unless the behavior can be 
;;; modelled in the latch model itself
;;;

(defclass wire ()
  ((signal-value :accessor signal-value :initarg :signal-value)
   ;; writing should only happen through set-signal, to 
   ;; trigger change-actions
   (change-actions :accessor change-actions :initarg :change-actions
                   :initform nil)))

(defmethod add-action ((w wire) func &optional (output-initializations nil))
  "OUTPUT-INITIALIZATIONS should be either NIL or a list of functions 
without arguments. 
   If NIL, the action FUNC is called once immediately, to initialize outputs. 
For sequential circuits, that may not make sense. If non-NIL, each function 
in the OUTPUT-INITIALIZATIONS list is called in lieu of calling FUNC."

  (push func (change-actions w))
  (if (null output-initializations)
    (funcall func) ;; must call to ensure outputs assume correct state
    ;; for initial inputs 
    (mapcar #'funcall output-initializations)))
 
(defmethod set-signal ((w wire) new-value)
  ;; (format t "Setting signal of ~A to ~D~%" w new-value)
  (when (not (eql new-value (signal-value w)))
    (setf (signal-value w) new-value)
    (mapcar #'funcall (change-actions w))))

(defmethod get-signal ((w wire))
  (signal-value w))

(defun make-wire ()
  (make-instance 'wire :signal-value 0))

;;;
;;; AGENDA
;;;
;;; This holds the sequence of events to be processed. 
;;; 
;;; Circuit elements should respond to input events by placing
;;; output events on the agenda.
;;;

(defclass agenda ()
  ((time :accessor agenda-time :initarg :agenda-time)
   (segments :accessor segments :initarg :segments
             :documentation "List of time segments")))

(defmethod empty ((a agenda))
  (null (segments a)))

(defmethod first-segment ((a agenda))
  (car (segments a)))

(defmethod rest-segments ((a agenda))
  (cdr (segments a)))

(defun make-agenda (time)
  (make-instance 'agenda
                 :agenda-time time
                 :segments nil))

(defmethod add-to-agenda ((a agenda) action time)
  (labels 
      ((belongs-before (segments)
         (or (null segments)
             (< time (segment-time (car segments)))))
       (make-new-time-segment (action time)
         (let ((q (make-queue)))
           (insert q action)
           (make-time-segment time q)))
       (add-to-segments! (segments)
         (if (= time (segment-time (car segments)))
             (insert (segment-queue (car segments)) action)
             (let ((r (cdr segments)))
               (if (belongs-before r)
                   (push (make-new-time-segment action time)
                         (cdr segments))
                   (add-to-segments! r))))))

    (let ((segments (segments a)))
      (if (belongs-before segments)
          (setf (segments a)
                (cons (make-new-time-segment action time)
                      segments))
          (add-to-segments! segments)))))

(defmethod remove-first-item ((a agenda))
  (let ((q (segment-queue (first-segment a))))
    (delete-front q)
    (when (empty q)
      (setf (segments a) (rest-segments a)))))

(defmethod first-item ((a agenda))
  (if (empty a)
      (error "Empty agenda")
      (let ((first-seg (first-segment a)))
        (setf (agenda-time a) (segment-time first-seg))
        (front (segment-queue first-seg)))))

;;; TIME-SEGMENT
;;;
;;; the element of an agenda
;;; contains a queue of actions, all of which occur
;;; at the same time.

(defclass time-segment ()
  ((time :accessor segment-time :initarg :segment-time)
   (queue :accessor segment-queue :initarg :segment-queue)))

(defun make-time-segment (time queue)
  (make-instance 'time-segment :segment-time time :segment-queue queue))

;;;
;;; PROPAGATION ACTIONS
;;;
;;; These are the routines that allow devices
;;; to propagate their inputs to their outputs
;;;

(defvar *agenda* (make-agenda 0))

(defun after-delay-fun (delay action)
  (add-to-agenda *agenda* action (+ delay (agenda-time *agenda*))))

(defun at-time-fun (time action)
  (add-to-agenda *agenda* action time))

(defmacro at-time (time &body actions)
  (let ((t-sym (gensym)))
    `(let ((,t-sym ,time))
       (add-to-agenda *agenda* #'(lambda () ,@actions)
		      ,t-sym))))

(defmacro after-delay (delay &body actions)
  `(add-to-agenda *agenda* #'(lambda () ,@actions)
		  (+ ,delay (agenda-time *agenda*))))

(defun propagate ()
  (do ((item))
      ((empty *agenda*) 'done)
    (setf item (first-item *agenda*))
    (funcall item)
    (remove-first-item *agenda*)))

;;;
;;; CIRCUIT DEFINITIONS
;;;

(defmethod probe ((w wire) name)
  (add-action w #'(lambda () (format t "~A Time: ~D New Value: ~A~%"
                                     name (agenda-time *agenda*)
                                    (get-signal w)))))
    

(defmacro accumulate-probe (place wire)
  "Install a probe on WIRE which pushes (time . new-value)
into PLACE"
  (let ((w (gensym)))
    `(let ((,w ,wire))
       (add-action ,w #'(lambda () 
				   (push (cons (agenda-time *agenda*)
					       (get-signal ,w))
					 ,place))))))

(defclass gated-probe ()
  ((enabled :initarg :enabled :accessor enabled)))

(defmethod make-gated-edge-probe ((w wire) (edge-wire wire) name
				  &optional (edge-type :rising))

  "Makes a gated probe which samples W when EDGE-WIRE has an edge transition
that matches EDGE-TYPE (:RISING or :FALLING). Should add setup & hold time
requirements on this probe, but for now, close enough."

  (let ((probe (make-instance 'gated-probe :enabled nil)))
    (add-action edge-wire
		(ecase edge-type
		  (:rising
		   #'(lambda ()
		       (when (logical-true (get-signal edge-wire))
			 (when (enabled probe)
			   (format t 
				   "~A Time: ~D New Value: ~A~%"
				   name (agenda-time *agenda*)
				   (get-signal w))))))
		   
		  (:falling
		   #'(lambda ()
		       (when (logical-false (get-signal edge-wire))
			 (when (enabled probe)
			   (format t 
				   "~A Time: ~D New Value: ~A~%"
				   name (agenda-time *agenda*)
				   (get-signal w))))))))
    probe))

#||
;;; SICP example
;;; see circuit-test::sicp-example

(defconstant +sicp-inv-delay+ 2)
(defconstant +sicp-and-delay+ 3)
(defconstant +sicp-or-delay+ 5)

(setf *in1* (make-wire))
(setf *in2* (make-wire))
(setf *sum* (make-wire))
(setf *carry* (make-wire))

(probe *sum* 'sum)
(probe *carry* 'carry)

(half-adder  *in1* *in2* *sum* *carry* +sicp-or-delay+ 
	     +sicp-and-delay+ +sicp-inv-delay+)
(set-signal *in1* 1)
(propagate)
(set-signal *in2* 1)
(propagate)

||#

                                                           
