;;;
;;; memory-components.lisp
;;;
;;; An attempt to model memories for circuit-sim.lisp
;;;
;;; $Revision: 1.2 $
;;;
;;; $Log: memory-components.lisp,v $
;;; Revision 1.2  2007/02/11 18:49:44  jao
;;; In package "CIRCUIT"
;;;
;;; Revision 1.1  2007/02/11 00:40:15  jao
;;; Initial revision
;;;
;;;

(in-package "CIRCUIT")

(defun make-rom (address-bus data-bus rom-contents propagation-delay &key 
		 (enable-pin nil) (enable-pin-active 0))
  "Uses address-bus values to index into rom-contents, putting the
rom contents onto data-bus. enable-pin, if non-nil, gates the activity
on data-bus. Unless the enable-pin matches enable-pin-active, no signal
is forced onto the data-bus."

  (labels ((read-data ()
	     (let ((a (get-integer-signal address-bus)))
	       ; (format t "Address ~D~%" a)
	       (if (numberp a)
		   (aref rom-contents a)
		   a)))
	   
	   (event-process ()
	     (if (or (not enable-pin)
		     (logical-eqv enable-pin-active
				  (get-signal enable-pin)))
		 (after-delay propagation-delay
			      ;;     (format t "data bus to ~D~%" 
			      ;;	      (read-data))
			      (set-signal data-bus 
					  (read-data))))))
    (add-action address-bus #'event-process)
    (when enable-pin
      (add-action enable-pin #'event-process))))
  
  ;;;
  ;;; neat feature would be if :invalid address-bits would 
  ;;; still produce valid output if it is a don't care. I.e. if 
  ;;; all combinations of 0 and 1 for the :invalid bits would still
  ;;; produce the same data value (or partial values? I.e. some data bits
  ;;; are all zero or all one, although some disagree)
  ;;; 
  ;;; need a function to produce "all possible valid combinations"
  ;;; e.g. (1 0 :invalid) would produce '((1 0 0) (1 0 1)),
  ;;; then you would test all these addresses to see if the data (or some 
  ;;; bits?) concurred.
  ;;;
  ;;; e.g. (concordance '((1 0 0) (1 0 1)) produces (1 0 :invalid))

#||
;;; ROM test

(setf *a0* (make-wire)
      *a1* (make-wire)
      *d0* (make-wire)
      *d1* (make-wire))
(setf *addbus* (make-bus *a0* *a1*))
(setf *databus* (make-bus *d0* *d1*))

(let ((data (make-array (list 4) :initial-contents
			'(3 2 1 0))))
  (make-rom *addbus* *databus* data 1))

(probe-integer *databus* 'data)
; (probe *d0* 'd0)
; (probe *d1* 'd1)

(set-signal *addbus* 0)
(propagate)
(set-signal *addbus* 1)
(propagate)
(set-signal *addbus* 2)
(propagate)
(set-signal *addbus* 3)
(propagate)

||#

(defun make-addressable-latch (addr-bus data-bit enable clear data-outputs
			       prop-delay-addr prop-delay-data 
			       prop-delay-enable prop-delay-clear
			       &key (setup-data 0) (setup-addr 0)
			       (hold-data 0) 
			       (min-clear-width 0) (min-enable-width 0))
  "Represents an addressable latch, such as the 74LS259. The addr-bus selects 
among the various one-bit latches making up the data-outputs. 

  clear enable |  addressed output    other outputs
  -------------|----------------------------------|
    H     L    |      data-bit        holds value | (set addressed latch)
    H     H    |      holds value     holds value | (memory)
    L     L    |      data-bit          L         | (1-N mux)
    L     H    |         L              L         | (clear latches)
  -------------------------------------------------

  setup and hold times are relative to the rising edge of the enable pulse.
SETUP AND HOLD NOT YET ENFORCED."

  (let ((now (agenda-time *agenda*)))
    (let ((last-data-in-change now)
	  (last-clear-falling now))
      
      (labels ((address ()
		 (get-integer-signal addr-bus))

	       (selected-bit ()
		 (nth-wire (address) data-outputs))
	       
	       (set-mode ()
		 (and (logical-true (get-signal clear))
		      (logical-false (get-signal enable))))

	       (memory-mode ()
		 (and (logical-true (get-signal clear))
		      (logical-true (get-signal enable))))
	       (mux-mode ()
		 (and (logical-false (get-signal clear))
		      (logical-false (get-signal enable))))

	       (clear-mode ()
		 (and (logical-false (get-signal clear))
		      (logical-true (get-signal enable))))
		 
	       (data-process () ;; events on data-bit input
		 (when (or (mux-mode) (set-mode))
		   ;; 1->N mux mode or set addressed latch
		   (after-delay prop-delay-data 
				(set-signal (selected-bit)
					    (get-signal 
					     data-bit)))))

	       (clear-process () ;; events on clear
		 (cond
		   ((clear-mode) ;; falling from memory to clear
		    (after-delay prop-delay-clear
				 (set-signal data-outputs 0)))

		   ((mux-mode) ;; falling from set-mode to mux
		    (let ((addr (address)))
		      (iterate-bus-counting 
		       #'(lambda (w i)
			   (unless (= i addr)
			     (after-delay prop-delay-clear
					  (set-signal w 0))))
		       data-outputs)))))

	       (enable-process () ;; events on enable
		 (cond 
		   ((or (mux-mode) (set-mode))
		    (after-delay prop-delay-enable
				 (set-signal (selected-bit)
					     (get-signal data-bit))))
		   ((clear-mode)
		    ;; if clear is low, go from 1->N mux to 
		    ;; clear latches, otherwise data-bit -> holds value
		    (after-delay prop-delay-enable
				 (set-signal (selected-bit)
					     0)))))

	       (addr-process () ;; process change of address
		 ;; matters when addressed output and 
		 ;; other outputs differ, i.e. in 
		 ;; set addressed latch and 1-->N mux mode
		 ;; (format t "Addr change to ~D~%" (address))
		 (cond
		   ((set-mode) ; set new output
		    ;;; (format t "Set mode~%")
		    (after-delay prop-delay-enable
				 (set-signal (selected-bit)
					     (get-signal data-bit)))
		    ;;; (format t "End set mode~%")
		    )
		   ((mux-mode) ; set addressed output, clear others
		    ;;; (format t "Mux mode: address change to ~D~%"
		    ;;;    (address))
		    (let ((addr (address)))
		      (iterate-bus-counting 
		       #'(lambda (w i)
			   (after-delay prop-delay-addr
					(set-signal w
						    (if (= i addr)
							(get-signal
							 data-bit)
							0))))
		       data-outputs))
		    ;; (format t "End mux mode~%")
		    )
		   ;; (t (format t "neither mode~%"))
		   )))
	       
	(add-action addr-bus #'addr-process)
	(add-action clear #'clear-process)
	(add-action enable #'enable-process)
	(add-action data-bit #'data-process)))))

#||

;;; test addressable latch
;;;
;;; 4-bit latch


(progn
  (setf *agenda* (make-agenda 0)
	*s0* (make-wire)
	*s1* (make-wire)
	*clear* (make-wire)
	*d* (make-wire)
	*enable* (make-wire)
	*q0* (make-wire)
	*q1* (make-wire)
	*q2* (make-wire)
	*q3* (make-wire))
  
  (setf *a-bus* (make-bus *s0* *s1*))
  (setf *d-bus* (make-bus *q0* *q1* *q2* *q3*))
  
  (make-addressable-latch *a-bus*
			  *d*
			  *enable*
			  *clear*
			  *d-bus*
			  1 1 1 1)
  (probe-integer *d-bus* 'data)
  
  (set-signal *a-bus* 0)
  (set-signal *d* 1)
  (set-signal *clear* nil)
  (set-signal *enable* nil)

  ;; mux-mode
  

  (at-time 0 
	   (format t "~%Sweeping 1 up through outputs.~%")
	   (set-signal *a-bus* 0))

  (at-time 2 (set-signal *a-bus* 1))
  (at-time 4 (set-signal *a-bus* 2))
  (at-time 6 (set-signal *a-bus* 3))

  (at-time 8 
	   (format t "input data going low.~%")
	   (set-signal *d* 0))
 
  (at-time 10 
	   (format t "Going to set-memory mode~%")
	   (set-signal *clear* t)) ;; go to memory

  (at-time 12 (set-signal *a-bus* 2)
	   (set-signal *d* 1))
  (at-time 14 (set-signal *a-bus* 1))
  (at-time 16 
	   (set-signal *d* 0)
	   (set-signal *a-bus* 0))
  
  (at-time 20 
	   (format t "going to hold-memory mode~%")
	   (set-signal *enable* t))

  (at-time 22 (set-signal *a-bus* 1))
  (at-time 24 (set-signal *a-bus* 2))
  (at-time 24 (set-signal *a-bus* 3))
  (at-time 26 (set-signal *a-bus* 4))
  (at-time 28 (set-signal *d* 1))
	   
  (at-time 30 (set-signal *a-bus* 3))
  (at-time 32 (set-signal *a-bus* 2))
  (at-time 34 (set-signal *a-bus* 1))
  (at-time 36 (set-signal *a-bus* 0))
  
  (at-time 38
	   (format t "Clearing register~%")
	   (set-signal *clear* 0))

  (at-time 40 (set-signal *d* 1))
  (at-time 42 (set-signal *a-bus* 1))
  (at-time 44 (set-signal *a-bus* 2))
  (at-time 46 (set-signal *a-bus* 3))
  
  (at-time 48 
	   (format t "Back to mux-mode~%")
	   (set-signal *enable* 0))

  (propagate))

||#

(defun make-shift-register-shared-IO 
    (data-io-bus clock clear enables function-s0 function-s1
     data-in-sl data-in-sr aux-output-msb aux-output-lsb
     prop-delay-clock prop-delay-enable 
     &key (setup-data 0) (setup-function 0) (setup-clear 0)
     (hold-data 0) (hold-function 0) (hold-clear 0))
  "Represents a shift register with shared data input and output lines,
such as the 74LS323. The data-io-bus is usually driven with the register
value, but when the function is 'load', the register outputs are 
put into tri-state mode, and the data-io-bus value is clocked into 
the register. The clock is active on the rising edge.

  If any of the enables are high, then the ouputs are not driven. However,
the register can still be loaded, shifted, or cleared.

  function-s0 and -s1 control the operation of the register.
  clear is active-low, and synchronously clears the register except in
load operation.

  CLR  S1  S0   CLK | 
 -------------------|------------------------
   L    X   X    ^  | clear: register to low
   H    L   H    ^  | shift right: data-in-SR goes into MSB of register
   H    H   L    ^  | shift left: data-in-SL goes into LSB of register
   H    L   L    X  | hold: data values maintained

for the previous four rows, unless all enables are low, the data is not
driven at the outputs. However, the aux-output-msb and -lsb always are
driven by the corresponding outputs of the flip-flops.
   
   H    H   H    ^  | synchronous load: values on the I/O lines are 
                    | loaded into the register.

  In the current implementation, setup and hold times are not enforced."

  (let* ((register-width (bus-width data-io-bus))
	 (register-values (make-array (list register-width))))

    (labels ((shift-right-mode ()
	       (and (logical-true (get-signal clear))
		    (logical-true (get-signal function-s0))
		    (logical-false (get-signal function-s1))))
	     (shift-left-mode ()
	       (and (logical-true (get-signal clear))
		    (logical-false (get-signal function-s0))
		    (logical-true (get-signal function-s1))))
	     (hold-mode ()
	       (and (logical-true (get-signal clear))
		    (logical-false (get-signal function-s0))
		    (logical-false (get-signal function-s1))))
	     (load-mode ()
	       (and (logical-true (get-signal clear))
		    (logical-true (get-signal function-s0))
		    (logical-true (get-signal function-s1))))
	     
	     (clear-mode ()
	       (logical-false (get-signal clear)))

	     (drive-output-mode ()
	       (and 
		(zerop (get-integer-signal enables))
		(not (and (logical-true (get-signal function-s0))
			  (logical-true (get-signal function-s1))))))
	     
	     (update-outputs ()
	       (set-signal aux-output-lsb
			   (aref register-values 0))
	       (set-signal aux-output-msb
			   (aref register-values (1- register-width)))
	       (when (drive-output-mode)
		 (iterate-bus-counting 
		  #'(lambda (w i)
		      (set-signal w (aref register-values i)))
		  data-io-bus)))
	     
	     (process-clock-rising ()
	       (cond

		 ((clear-mode)
		  (dotimes (i register-width)
		    (setf (aref register-values i) 0)))

		 ((load-mode)
		  (iterate-bus-counting
		   #'(lambda (w i)
		       (setf (aref register-values i)
			     (get-signal w)))
		   data-io-bus))

		 ((shift-left-mode)
		  (let ((lsb (get-signal data-in-sl)))
;; 		     (format t "~S: shifting LEFT in ~D to LSB~%"
;;		        (reverse register-values) lsb)
		    (dotimes (i (1- register-width))
		      (setf (aref register-values 
				  (- register-width (1+ i)))
			    (aref register-values 
				  (- register-width (+ 2 i)))))
		    (setf (aref register-values 0) lsb)
;;		     (format t "after: ~S~%" (reverse register-values))
		    ))

		 ((shift-right-mode)
		  (let ((msb (get-signal data-in-sr)))
;;		     (format t "~S: shifting RIGHT in ~D to MSB~%"
;;		     (reverse register-values) msb)
		    (dotimes (i (1- register-width))
		      (setf (aref register-values i)
			    (aref register-values (1+ i))))
		    (setf (aref register-values (1- register-width))
			  msb)
;;		     (format t "after: ~S~%" (reverse register-values))
		    ))
		    
		    ((hold-mode))

		 (t ; unknown-mode
		  (error "shift register clocked in unknown mode at time ~D."
			 (agenda-time *agenda*))))

	       (after-delay prop-delay-clock
			    (update-outputs)))

	     (process-clock ()
	       (when (logical-true (get-signal clock))
		 (process-clock-rising)))
	     
	     (process-enables ()
	       (when (drive-output-mode)
		 (after-delay prop-delay-enable
			      (update-outputs)))))
      
      (add-action clock #'process-clock)
      (add-action enables #'process-enables))))

#||
;;; test shift register
;;; 
;;; 4-bit shift register
;;;

(progn
  (setf *agenda* (make-agenda 0)
	*s0* (make-wire)
	*s1* (make-wire)
	*clear* (make-wire)
	*clock* (make-wire)
	*enable* (make-wire)
	*d0* (make-wire)
	*d1* (make-wire)
	*d2* (make-wire)
	*d3* (make-wire)
	*q0* (make-wire)
	*q3* (make-wire))

  (setf *d-bus* (make-bus *d0* *d1* *d2* *d3*))
  
  (make-shift-register-shared-io
   *d-bus* *clock* *clear* *enable* *s0* *s1*
   *q3* *q0* ; data in SL, data in SR
   *q3* *q0* ; MSB LSB
   1 1)

  ;; circular connections...

  (probe-integer *d-bus* 'data)
  (make-clock *clock* 1 :rising 2 2) ; rising edges at 1, 5, 9, 13,
  (probe *clock* 'clock)

  (set-signal *clear* 0)
  (set-signal *enable* 0)
  (set-signal *s0* 1)
  (set-signal *s1* 1)

  (at-time 3 
	   (format t "~%Initial load~%")
	   (set-signal *d-bus* #xA)
	   (set-signal *clear* 1)) ; edge at 5 loads

  (at-time 6
	   (format t "Holding~%")
	   (set-signal *s0* 0)
	   (set-signal *s1* 0)) ; edge at 9, 13 hold

  (at-time 14
	   (format t "Rotating to left~%")
	   (set-signal *s1* 1))

  ; rotate at 17, 21, 25, 29

  (at-time 30
	   (format t "Loading new data~%")
	   (set-signal *s0* 1)
	   (set-signal *s1* 1)
	   (set-signal *d-bus* 3))

  ; load at 33
  
  (at-time 34
	   (format t "Rotating to right~%")
	   (set-signal *s0* 1)
	   (set-signal *s1* 0))

  ; shift right at 37, 41, 45, 49, 53

  (propagate-until 55))

||#
